/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	Jacob's Matrix Development
 *
 * NAME:        main.cpp ( Jacob's Matrix, C++ )
 *
 * COMMENTS:
 */

#include <libtcod.hpp>
#include <stdio.h>
#include <time.h>

//#define USE_AUDIO

#define POPUP_DELAY 1000

#ifdef WIN32
#include <windows.h>
#include <process.h>
#define usleep(x) Sleep((x)*1000)
#pragma comment(lib, "SDL.lib")
#pragma comment(lib, "SDLmain.lib")

// I really hate windows.
#ifdef min
#undef min
#endif
#ifdef max
#undef max
#endif

#else
#include <unistd.h>
#endif

#include <fstream>
using namespace std;

int		 glbLevelStartMS = -1;

#include "rand.h"
#include "gfxengine.h"
#include "config.h"

#include <SDL.h>
#ifdef USE_AUDIO
#include <SDL_mixer.h>

Mix_Music		*glbTunes = 0;
#endif
bool			 glbMusicActive = false;

// Cannot be bool!
// libtcod does not accept bools properly.

volatile int		glbItemCache = -1;

SPELL_NAMES		glbLastSpell = SPELL_NONE;
int			glbLastItemIdx = 0;
int			glbLastItemAction = 0;

void endOfMusic()
{
    // Keep playing.
    // glbMusicActive = false;
}

void
setMusicVolume(int volume)
{
#ifdef USE_AUDIO
    volume = BOUND(volume, 0, 10);

    glbConfig->myMusicVolume = volume;

    Mix_VolumeMusic((MIX_MAX_VOLUME * volume) / 10);
#endif
}

void
stopMusic(bool atoncedamnit = false)
{
#ifdef USE_AUDIO
    if (glbMusicActive)
    {
	if (atoncedamnit)
	    Mix_HaltMusic();
	else
	    Mix_FadeOutMusic(2000);
	glbMusicActive = false;
    }
#endif
}

void
startMusic()
{
#ifdef USE_AUDIO
    if (glbMusicActive)
	stopMusic();

    if (!glbTunes)
	glbTunes = Mix_LoadMUS(glbConfig->musicFile());
    if (!glbTunes)
    {
	printf("Failed to load %s, error %s\n", 
		glbConfig->musicFile(), 
		Mix_GetError());
    }
    else
    {
	glbMusicActive = true;
	glbLevelStartMS = TCOD_sys_elapsed_milli();

	if (glbConfig->musicEnable())
	{
	    Mix_PlayMusic(glbTunes, -1);		// Infinite
	    Mix_HookMusicFinished(endOfMusic);

	    setMusicVolume(glbConfig->myMusicVolume);
	}
    }
#endif
}


#ifdef main
#undef main
#endif

#include "mob.h"
#include "map.h"
#include "text.h"
#include "display.h"
#include "engine.h"
#include "panel.h"
#include "msg.h"
#include "firefly.h"
#include "item.h"
#include "chooser.h"

PANEL		*glbMessagePanel = 0;
DISPLAY		*glbDisplay = 0;
MAP		*glbMap = 0;
ENGINE		*glbEngine = 0;
PANEL		*glbPopUp = 0;
PANEL		*glbJacob = 0;
PANEL		*glbWeaponInfo = 0;
PANEL		*glbVictoryInfo = 0;
bool		 glbPopUpActive = false;
bool		 glbRoleActive = false;
FIREFLY		*glbHealthFire = 0;
FIREFLY		*glbManaFire = 0;
FIREFLY		*glbCoinStack = 0;
int		 glbLevel = 1;
int		 glbMaxLevel = 0;
PTRLIST<int>	 glbLevelTimes;
bool		 glbVeryFirstRun = true;

CHOOSER		*glbLevelBuildStatus = 0;

CHOOSER		*glbChooser = 0;
bool		 glbChooserActive = false;

int		 glbInvertX = -1, glbInvertY = -1;
int		 glbMeditateX = -1, glbMeditateY = -1;

#define DEATHTIME 15000


BUF
mstotime(int ms)
{
    int		sec, min;
    BUF		buf;

    sec = ms / 1000;
    ms -= sec * 1000;
    min = sec / 60;
    sec -= min * 60;

    if (min)
	buf.sprintf("%dm%d.%03ds", min, sec, ms);
    else if (sec)
	buf.sprintf("%d.%03ds", sec, ms);
    else
	buf.sprintf("0.%03ds", ms);

    return buf;
}

class MOBPTR_SORT
{
public:
    bool operator()(MOB *a, MOB *b)
    {
	return a->getUID() < b->getUID();
    }
};

class ITEMPTR_SORT
{
public:
    bool operator()(ITEM *a, ITEM *b)
    {
	return a->getUID() < b->getUID();
    }
};

class ITEMPTR_QUALITYSORT
{
public:
    bool operator()(ITEM *a, ITEM *b)
    {
	if (a->isBroken() && !b->isBroken())
	    return true;
	if (!a->isBroken() && b->isBroken())
	    return false;
	if (a->isMelee())
	{
	    if (b->isArmour())
		return true;
	    if (b->isMelee())
		return (a->getKillSpeed() < b->getKillSpeed());
	    return true;
	}
	if (a->isArmour())
	{
	    if (b->isMelee())
		return false;
	    if (b->isArmour())
		return (a->getAC() < b->getAC());
	    return true;
	}
	return a->getUID() < b->getUID();
    }
};

// Should call this in all of our loops.
void
redrawWorldNoFlush()
{
    static MAP	*centermap = 0;
    static POS	 mapcenter;
    bool	 isblind = false;
    bool	 isvictory = false;

    gfx_updatepulsetime();

    // Wipe the world
    for (int y = 0; y < SCR_HEIGHT; y++)
	for (int x = 0; x < SCR_WIDTH; x++)
	    gfx_printchar(x, y, ' ');

    if (glbMap)
	glbMap->decRef();
    glbMap = glbEngine ? glbEngine->copyMap() : 0;

    glbWeaponInfo->clear();

    glbMeditateX = -1;
    glbMeditateY = -1;
    int			avataralive = 0;
    if (glbMap && glbMap->avatar())
    {
	MOB		*avatar = glbMap->avatar();
	int		 timems;

	isvictory = glbMap->getNumOrcs() == 0;

	avataralive = avatar->alive();

	if (avatar->alive() && avatar->isMeditating())
	{
	    glbMeditateX = glbDisplay->x() + glbDisplay->width()/2;
	    glbMeditateY = glbDisplay->y() + glbDisplay->height()/2;
	}

	timems = TCOD_sys_elapsed_milli();

	if (centermap)
	    centermap->decRef();
	centermap = glbMap;
	centermap->incRef();
	mapcenter = avatar->meditatePos();

	isblind = avatar->hasItem(ITEM_BLIND);

	if (avatar->hasItem(ITEM_INVULNERABLE))
	    glbHealthFire->setFireType(FIRE_MONO);
	else if (avatar->hasItem(ITEM_POISON))
	    glbHealthFire->setFireType(FIRE_POISON);
	else
	    glbHealthFire->setFireType(FIRE_HEALTH);

	glbHealthFire->setRatioLiving(avatar->getHP() / (float) avatar->defn().max_hp);
	glbManaFire->setRatioLiving(avatar->getMP() / (float) avatar->defn().max_mp);
	{
	    int		ngold = 0;
	    ITEM	*gold = avatar->lookupItem(ITEM_GOLD);
	    if (gold)
		ngold = gold->getStackCount();

	    glbCoinStack->setParticleCount(BOUND(ngold, 0, 22 * 50));
	    glbCoinStack->setBloodCoins(5*avatar->map()->getDeadAdventurers());
	    glbCoinStack->setRatioLiving((float) (BOUND(ngold, 0, 20*48) / (20*48.)));
	}

	BUF		buf;

	{
	    MOBLIST	atlist;
	    glbMap->getAllMobsOfType(atlist, MOB_ADVENTURER);

	    glbWeaponInfo->setTextAttr(ATTR_WHITE);
	    glbWeaponInfo->appendText("* Adventurers *");
	    glbWeaponInfo->newLine();

	    glbWeaponInfo->setTextAttr(ATTR_NORMAL);
	    buf.sprintf("Retired: %d\nDead: %d\n",
		    glbMap->getRetiredAdventurers(),
		    glbMap->getDeadAdventurers());
	    glbWeaponInfo->appendText(buf);

	    MOBPTR_SORT	op;

	    atlist.stableSort(op);

	    if (!atlist.entries())
	    {
		glbWeaponInfo->setTextAttr(ATTR_NORMAL);
		glbWeaponInfo->appendText("  None Active");
		glbWeaponInfo->newLine();
	    }
	    for (int i = 0; i < atlist.entries(); i++)
	    {
		glbWeaponInfo->setTextAttr(ATTR_NORMAL);
		if (atlist(i)->pos().valid())
		    buf.sprintf("  %s", atlist(i)->getName());
		else
		    buf.sprintf("> %s", atlist(i)->getName());
		glbWeaponInfo->appendText(buf);
		glbWeaponInfo->newLine();

		buf.sprintf("   %d gold, %d kills",
			    atlist(i)->getGold(),
			    atlist(i)->getKills());
		glbWeaponInfo->appendText(buf);
		glbWeaponInfo->newLine();
	    }
	}
    }
    
    glbDisplay->display(mapcenter, isblind);
    glbMessagePanel->redraw();

    {
	glbJacob->clear();
	int		norcs = glbMap ? glbMap->getNumOrcs() : 0;

	DIFFICULTY_NAMES	diff = glbMap ? glbMap->getDifficulty() : DIFFICULTY_NORMAL;

	glbJacob->setTextAttr(ATTR_WHITE);
	if (norcs > glb_difficultydefs[diff].norc_yellow)
	    glbJacob->setTextAttr(ATTR_YELLOW);
	if (norcs > glb_difficultydefs[diff].norc_orange)
	    glbJacob->setTextAttr(ATTR_ORANGE);
	if (norcs > glb_difficultydefs[diff].norc_red)
	    glbJacob->setTextAttr(ATTR_RED);
	if (norcs < glb_difficultydefs[diff].norc_green)
	    glbJacob->setTextAttr(ATTR_GREEN);

	BUF		buf;

	buf.sprintf("Orcs: %d", 
		    norcs);

	glbJacob->appendText(buf);

	if (glbMap)
	{
	    glbJacob->setTextAttr(ATTR_NORMAL);
	    glbJacob->appendText("\n");
	    if (glbMap->getDifficulty() != DIFFICULTY_NORMAL)
	    {
		glbJacob->appendText("  ");
		glbJacob->appendText(glb_difficultydefs[glbMap->getDifficulty()].name);
	    }
	}
	

	if (glbMap && glbMap->avatar())
	{
	    MOB	*avatar = glbMap->avatar();

	    glbJacob->setTextAttr(ATTR_WHITE);
	    glbJacob->appendText("\nInventory:\n");

	    glbJacob->setTextAttr(ATTR_NORMAL);
	    if (!avatar->inventory().entries())
	    {
		glbJacob->appendText("  Nothing\n");
	    }
	    ITEMLIST	items = avatar->inventory();

	    ITEMPTR_SORT	op;

	    items.stableSort(op);

	    for (int i = 0; i < items.entries(); i++)
	    {
		BUF	itemname = items(i)->getName();
		itemname = gram_capitalize(itemname);
		buf.sprintf("  %s\n", itemname.buffer());
		glbJacob->appendText(buf);
	    }
	}
    }

    glbJacob->redraw();

    glbWeaponInfo->redraw();

    if (isvictory)
	glbVictoryInfo->redraw();

    if (glbCoinStack)
    {
	FIRETEX		*tex = glbCoinStack->getTex();
	tex->redraw(0, 0);
	tex->decRef();
    }
    {
	// we have to clear the RHS.
	static FIRETEX		 *cleartex = 0;
	if (!cleartex)
	{
	    cleartex = new FIRETEX(20, 48);
	    cleartex->buildFromConstant(false, 0, FIRE_HEALTH);
	}
	cleartex->redraw(60, 2);
    }

    if (glbMeditateX >= 0 && glbMeditateY >= 0)
    {
	gfx_printattrback(glbMeditateX, glbMeditateY, ATTR_AVATARMEDITATE);
    }

    if (glbInvertX >= 0 && glbInvertY >= 0)
    {
	gfx_printattr(glbInvertX, glbInvertY, ATTR_HILITE);
    }


    if (glbItemCache >= 0)
    {
	glbLevelBuildStatus->clear();

	glbLevelBuildStatus->appendChoice("   Building Distance Caches   ", glbItemCache);
	glbLevelBuildStatus->redraw();
    }

    if (glbChooserActive)
    {
	glbChooser->redraw();
    }

    if (glbPopUpActive)
	glbPopUp->redraw();
}

void
redrawWorld()
{
    redrawWorldNoFlush();
    TCODConsole::flush();
}

// Pops up the given text onto the screen.
// The key used to dismiss the text is the return code.
int
popupText(const char *buf, int delayms = 0)
{
    int		key = 0;
    int		startms = TCOD_sys_elapsed_milli();

    glbPopUp->clear();
    glbPopUp->resize(50, 48);
    glbPopUp->move(15, 1);

    // A header matches the footer.
    glbPopUp->appendText("\n");

    // Must make active first because append text may trigger a 
    // redraw if it is long!
    glbPopUpActive = true;
    glbPopUp->appendText(buf);

    // In case we drew due to too long text
    glbPopUp->erase();

    // Determine the size of the popup and resize.
    glbPopUp->shrinkToFit();

    while (!TCODConsole::isWindowClosed())
    {
	redrawWorld();
	key = gfx_getKey(false);
	if (delayms)
	{
	    int		currentms = (int)TCOD_sys_elapsed_milli() - startms;
	    float	ratio = currentms / (float)delayms;
	    ratio = BOUND(ratio, 0, 1);

	    glbPopUp->fillRect(0, glbPopUp->height()-1, (int)(glbPopUp->width()*ratio+0.5), 1, ATTR_WAITBAR);
	}
	if (key)
	{
	    // Don't let them abort too soon.
	    if (((int)TCOD_sys_elapsed_milli() - startms) >= delayms)
		break;
	}
    }

    glbPopUpActive = false;
    glbPopUp->erase();

    glbPopUp->resize(50, 30);
    glbPopUp->move(15, 10);

    redrawWorld();

    return key;
}

int
popupText(BUF buf, int delayms = 0)
{
    return popupText(buf.buffer(), delayms);
}

// Pops up the given text onto the screen.
// Allows the user to enter a response.
// Returns what they typed.
BUF
popupQuestion(const char *buf)
{
#define RESPONSE_LEN 50
    char	response[RESPONSE_LEN];

    glbPopUp->clear();
    glbPopUp->resize(50, 48);
    glbPopUp->move(15, 1);

    // A header matches the footer.
    glbPopUp->appendText("\n");

    // Must make active first because append text may trigger a 
    // redraw if it is long!
    glbPopUpActive = true;
    glbPopUp->appendText(buf);
    glbPopUp->appendText("\n");

    // In case we drew due to too long text
    glbPopUp->erase();

    // Determine the size of the popup and resize.
    glbPopUp->shrinkToFit();
    glbPopUp->resize( MAX(RESPONSE_LEN + 2, glbPopUp->getW()),
		      glbPopUp->getH() + 3 );

    redrawWorld();
    glbPopUp->getString("\n> ", response, RESPONSE_LEN);

    BUF		result;
    result.strcpy(response);

    glbPopUpActive = false;
    glbPopUp->erase();

    glbPopUp->resize(50, 30);
    glbPopUp->move(15, 10);

    redrawWorld();

    return result;
}

BUF
popupQuestion(BUF buf)
{
    return popupQuestion(buf.buffer());
}

template <typename T>
T
runMenuSelection(const PTRLIST<T> &list, T fallback, 
		 const PTRLIST<int> &hotkeys)
{
    // Run the chooser...
    if (!list.entries())
	return fallback;

    int		key;
    while (!TCODConsole::isWindowClosed())
    {
	redrawWorld();
	key = gfx_getKey(false);

	glbChooser->processKey(key);

	if (key)
	{
	    int		itemno = glbChooser->getChoice();
	    // User selects this?
	    if (key == '\n' || key == ' ')
	    {
		fallback = list(itemno);
		break;
	    }
	    else if (key == '\x1b')
	    {
		// Esc on options is to go back to play.
		// Play...
		break;
	    }
	    else
	    {
		// Ignore other options.
		for (int i = 0; i < hotkeys.entries(); i++)
		{
		    if (key == hotkeys(i))
		    {
			fallback = list(i);
			glbChooserActive = false;
			return fallback;
		    }
		}
	    }
	}
    }
    glbChooserActive = false;
    return fallback;
}

//
// Freezes the game until a key is pressed.
// If key is not direction, false returned and it is eaten.
// else true returned with dx/dy set.  Can be 0,0.
bool
awaitDirection(int &dx, int &dy)
{
    int			key;

    while (!TCODConsole::isWindowClosed())
    {
	redrawWorld();

	key = gfx_getKey(false);

	if (gfx_cookDir(key, dx, dy))
	{
	    return true;
	}
	if (key)
	    return false;
    }

    return false;
}

void
doExamine()
{
    int		x = glbDisplay->x() + glbDisplay->width()/2;
    int		y = glbDisplay->y() + glbDisplay->height()/2;
    int		key;
    int		dx, dy;
    POS		p;
    BUF		buf;
    bool	blind;

    glbInvertX = x;
    glbInvertY = y;
    blind = false;
    if (glbMap && glbMap->avatar())
	blind = glbMap->avatar()->hasItem(ITEM_BLIND);
    while (!TCODConsole::isWindowClosed())
    {
	redrawWorld();

	key = gfx_getKey(false);

	if (gfx_cookDir(key, dx, dy))
	{
	    // Space we want to select.
	    if (!dx && !dy)
		break;
	    x += dx;
	    y += dy;
	    x = BOUND(x, glbDisplay->x(), glbDisplay->x()+glbDisplay->width()-1);
	    y = BOUND(y, glbDisplay->y(), glbDisplay->y()+glbDisplay->height()-1);

	    glbInvertX = x;
	    glbInvertY = y;
	    p = glbDisplay->lookup(x, y);

	    if (p.tile() != TILE_INVALID && p.isFOV())
	    {
		p.describeSquare(blind);
	    }

	    msg_newturn();
	}

	// Any other key stops the look.
	if (key)
	    break;
    }

    // Check if we left on a mob.
    // Must re look up as displayWorld may have updated the map.
    p = glbDisplay->lookup(x, y);
    if (p.mob())
    {
	popupText(p.mob()->getLongDescription());
    }
    else if (p.item())
    {
	popupText(p.item()->getLongDescription());
    }

    glbInvertX = -1;
    glbInvertY = -1;
}

bool
doThrow(int itemno)
{
    int			dx, dy;

    msg_report("Throw in which direction?  ");
    if (awaitDirection(dx, dy) && (dx || dy))
    {
	msg_report(rand_dirtoname(dx, dy));
	msg_newturn();
    }
    else
    {
	msg_report("Cancelled.  ");
	msg_newturn();
	return false;
    }

    // Do the throw
    glbEngine->queue().append(COMMAND(ACTION_THROW, dx, dy, itemno));

    return true;
}

void
buildAction(ITEM *item, ACTION_NAMES *actions)
{
    glbChooser->clear();

    glbChooser->setTextAttr(ATTR_NORMAL);
    int			choice = 0;
    ACTION_NAMES	*origaction = actions;
    if (!item)
    {
	glbChooser->appendChoice("Nothing to do with nothing.");
	*actions++ = ACTION_NONE;
    }
    else
    {
	glbChooser->appendChoice("[ ] Do nothing");
	if (glbLastItemAction == ACTION_NONE)
	    choice = (int)(actions - origaction);
	*actions++ = ACTION_NONE;

	glbChooser->appendChoice("[X] Examine");
	if (glbLastItemAction == ACTION_EXAMINE)
	    choice = (int)(actions - origaction);
	*actions++ = ACTION_EXAMINE;

	if (item->isRanged())
	{
	    glbChooser->appendChoice("[B] Break");
	    if (glbLastItemAction == ACTION_BREAK)
		choice = (int)(actions - origaction);
	    *actions++ = ACTION_BREAK;
	}
	if (item->isPotion())
	{
	    glbChooser->appendChoice("[Q] Quaff");
	    if (glbLastItemAction == ACTION_QUAFF)
		choice = (int)(actions - origaction);
	    *actions++ = ACTION_QUAFF;
	    glbChooser->appendChoice("[T] Throw");
	    if (glbLastItemAction == ACTION_THROW)
		choice = (int)(actions - origaction);
	    *actions++ = ACTION_THROW;
	}
	if (item->isFood())
	{
	    glbChooser->appendChoice("[E] Eat");
	    if (glbLastItemAction == ACTION_EAT)
		choice = (int)(actions - origaction);
	    *actions++ = ACTION_EAT;
	}
	if (item->isRing())
	{
	    glbChooser->appendChoice("[W] Wear");
	    if (glbLastItemAction == ACTION_WEAR)
		choice = (int)(actions - origaction);
	    *actions++ = ACTION_WEAR;
	}
	if (!item->defn().isflag)
	{
	    glbChooser->appendChoice("[D] Drop");
	    if (glbLastItemAction == ACTION_DROP)
		choice = (int)(actions - origaction);
	    *actions++ = ACTION_DROP;
	}
    }
    glbChooser->setChoice(choice);

    glbChooserActive = true;
}

bool
useItem(MOB *mob, ITEM *item, int itemno)
{
    ACTION_NAMES		actions[NUM_ACTIONS+1];
    bool			done = false;
    int				key;

    buildAction(item, actions);
    glbChooserActive = true;

    // Run the chooser...
    while (!TCODConsole::isWindowClosed())
    {
	redrawWorld();
	key = gfx_getKey(false);

	glbChooser->processKey(key);

	if (key)
	{
	    if (key == 'd' || key == 'D')
	    {
		glbEngine->queue().append(COMMAND(ACTION_DROP, itemno));
		done = true;
		break;
	    }
	    else if (key == 'q' || key == 'Q')
	    {
		glbEngine->queue().append(COMMAND(ACTION_QUAFF, itemno));
		done = true;
		break;
	    }
	    else if (key == 'e' || key == 'E')
	    {
		glbEngine->queue().append(COMMAND(ACTION_EAT, itemno));
		done = true;
		break;
	    }
	    else if (key == 'b' || key == 'B')
	    {
		glbEngine->queue().append(COMMAND(ACTION_BREAK, itemno));
		done = true;
		break;
	    }
	    else if (key == 'w' || key == 'W')
	    {
		glbEngine->queue().append(COMMAND(ACTION_WEAR, itemno));
		done = true;
		break;
	    }
	    else if (key == 'x' || key == 'X')
	    {
		popupText(item->getLongDescription());
		done = false;
		break;
	    }
	    else if (key == 't' || key == 'T')
	    {
		glbChooserActive = false;
		doThrow(itemno);
		done = true;
		break;
	    }
	    // User selects this?
	    else if (key == '\n' || key == ' ')
	    {
		ACTION_NAMES	action;

		action = actions[glbChooser->getChoice()];
		glbLastItemAction = action;
		switch (action)
		{
		    case ACTION_DROP:
			glbEngine->queue().append(COMMAND(ACTION_DROP, itemno));
			done = true;
			break;
		    case ACTION_QUAFF:
			glbEngine->queue().append(COMMAND(ACTION_QUAFF, itemno));
			done = true;
			break;
		    case ACTION_BREAK:
			glbEngine->queue().append(COMMAND(ACTION_BREAK, itemno));
			done = true;
			break;
		    case ACTION_WEAR:
			glbEngine->queue().append(COMMAND(ACTION_WEAR, itemno));
			done = true;
			break;
		    case ACTION_EAT:
			glbEngine->queue().append(COMMAND(ACTION_EAT, itemno));
			done = true;
			break;
		    case ACTION_SEARCH:
			glbEngine->queue().append(COMMAND(ACTION_SEARCH, itemno));
			done = true;
			break;
		    case ACTION_MEDITATE:
			glbEngine->queue().append(COMMAND(ACTION_MEDITATE, itemno));
			done = true;
			break;
		    case ACTION_EXAMINE:
			popupText(item->getLongDescription());
			done = false;
			break;
		    case ACTION_THROW:
			glbChooserActive = false;
			doThrow(itemno);
			done = true;
			break;
		}
		break;
	    }
	    else if (key == '\x1b')
	    {
		// Esc on options is to go back to play.
		// Play...
		break;
	    }
	    else
	    {
		// Ignore other options.
	    }
	}
    }
    glbChooserActive = false;

    return done;
}

void
buildInventory(MOB *mob)
{
    glbChooser->clear();

    glbChooser->setTextAttr(ATTR_NORMAL);
    if (!mob)
    {
	glbChooser->appendChoice("The dead own nothing.");
    }
    else
    {
	if (!mob->inventory().entries())
	{
	    glbChooser->appendChoice("nothing");
	}
	else
	{
	    for (int i = 0; i < mob->inventory().entries(); i++)
	    {
		ITEM 		*item = mob->inventory()(i);
		BUF		 name = item->getName();
		if (item->isEquipped())
		{
		    BUF		equip;
		    equip.sprintf("%s (equipped)", name.buffer());
		    name.strcpy(equip);
		}

		glbChooser->appendChoice(name);
	    }
	}
    }
    glbChooser->setChoice(glbLastItemIdx);

    glbChooserActive = true;
}

void
inventoryMenu()
{
    MOB		*avatar = (glbMap ? glbMap->avatar() : 0);
    int		key;
    bool	setlast = false;
    buildInventory(avatar);

    // Run the chooser...
    while (!TCODConsole::isWindowClosed())
    {
	redrawWorld();
	key = gfx_getKey(false);

	glbChooser->processKey(key);

	if (key)
	{
	    int		itemno = glbChooser->getChoice();
	    bool	done = false;
	    // User selects this?
	    if (key == '\n' || key == ' ')
	    {
		done = true;

		glbLastItemIdx = glbChooser->getChoice();
		setlast = true;
		if (itemno >= 0 && itemno < avatar->inventory().entries())
		{
		    done = useItem(avatar, avatar->inventory()(itemno), itemno);
		}
		// Finish the inventory menu.
		if (done)
		{
		    break;
		}
		else
		{
		    buildInventory(avatar);
		    setlast = false;
		}
	    }
	    else if (key == '\x1b' || key == 'i')
	    {
		// Esc on options is to go back to play.
		// Play...
		break;
	    }
	    else if (itemno >= 0 && itemno < avatar->inventory().entries())
	    {
		if (key == 'd' || key == 'D')
		{
		    glbEngine->queue().append(COMMAND(ACTION_DROP, itemno));
		    done = true;
		}
		else if (key == 'q' || key == 'Q')
		{
		    glbEngine->queue().append(COMMAND(ACTION_QUAFF, itemno));
		    done = true;
		}
		else if (key == 'e' || key == 'E')
		{
		    glbEngine->queue().append(COMMAND(ACTION_EAT, itemno));
		    done = true;
		}
		else if (key == 'b' || key == 'B')
		{
		    glbEngine->queue().append(COMMAND(ACTION_BREAK, itemno));
		    done = true;
		}
		else if (key == 'w' || key == 'W')
		{
		    glbEngine->queue().append(COMMAND(ACTION_WEAR, itemno));
		    done = true;
		}
		else if (key == 'x' || key == 'X')
		{
		    if (itemno >= 0 && itemno < avatar->inventory().entries())
		    {
			popupText(avatar->inventory()(itemno)->getLongDescription());
		    }
		}
		else if (key == 't' || key == 'T')
		{
		    glbLastItemIdx = glbChooser->getChoice();
		    setlast = true;
		    glbChooserActive = false;
		    doThrow(itemno);
		    done = true;
		    break;
		}
	    }
	    else
	    {
		// Ignore other options.
	    }

	    if (done)
		break;
	}
    }
    if (!setlast)
	glbLastItemIdx = glbChooser->getChoice();
    glbChooserActive = false;
}

void
castSpell(MOB *mob, SPELL_NAMES spell)
{
    int			dx = 0;
    int			dy = 0;

    if (glb_spelldefs[spell].needsdir)
    {
	msg_report("Cast in which direction?  ");
	if (awaitDirection(dx, dy) && (dx || dy))
	{
	    msg_report(rand_dirtoname(dx, dy));
	    msg_newturn();
	}
	else
	{
	    msg_report("Cancelled.  ");
	    msg_newturn();
	    return;
	}
    }
    glbLastSpell = spell;
    glbEngine->queue().append(COMMAND(ACTION_CAST, dx, dy, spell));
}

void
buildZapList(MOB *mob, PTRLIST<SPELL_NAMES> &list)
{
    glbChooser->clear();
    list.clear();
    int				chosen = 0;

    glbChooser->setTextAttr(ATTR_NORMAL);
    if (!mob)
    {
	glbChooser->appendChoice("The dead can cast nothing.");
	list.append(SPELL_NONE);
    }
    else
    {
	SPELL_NAMES		spell;

	FOREACH_SPELL(spell)
	{
	    ITEM_NAMES		itemname;
	    itemname = (ITEM_NAMES) glb_spelldefs[spell].item;
	    if (itemname != ITEM_NONE &&
		mob->hasItem(itemname))
	    {
		BUF		buf;

		buf.sprintf("Cast %s", glb_spelldefs[spell].name);
		glbChooser->appendChoice(buf);
		list.append(spell);
		if (spell == glbLastSpell)
		    chosen = list.entries()-1;
	    }
	}
    }
    if (!glbChooser->entries())
    {
	glbChooser->appendChoice("You know no magic spells.");
	list.append(SPELL_NONE);
    }

    glbChooser->setChoice(chosen);

    glbChooserActive = true;
}

void
zapMenu()
{
    PTRLIST<SPELL_NAMES>	spelllist;
    PTRLIST<int>		hotkeys;
    MOB		*avatar = (glbMap ? glbMap->avatar() : 0);

    buildZapList(avatar, spelllist);

    SPELL_NAMES	spell = runMenuSelection(spelllist, SPELL_NONE, hotkeys);

    if (spell != SPELL_NONE)
    {
	castSpell(avatar, spell);
    }
    glbChooserActive = false;
}

void
buildOptionsMenu(OPTION_NAMES d)
{
    OPTION_NAMES	option;
    glbChooser->clear();

    glbChooser->setTextAttr(ATTR_NORMAL);
    FOREACH_OPTION(option)
    {
	glbChooser->appendChoice(glb_optiondefs[option].name);
    }
    glbChooser->setChoice(d);

    glbChooserActive = true;
}

bool
optionsMenu()
{
    int			key;
    bool		done = false;

    buildOptionsMenu(OPTION_INSTRUCTIONS);

    // Run the chooser...
    while (!TCODConsole::isWindowClosed())
    {
	redrawWorld();
	key = gfx_getKey(false);

	glbChooser->processKey(key);

	if (key)
	{
	    // User selects this?
	    if (key == '\n' || key == ' ')
	    {
		if (glbChooser->getChoice() == OPTION_INSTRUCTIONS)
		{
		    // Instructions.
		    popupText(text_lookup("game", "help"));
		}
		else if (glbChooser->getChoice() == OPTION_PLAY)
		{
		    // Play...
		    break;
		}
#if 0
		else if (glbChooser->getChoice() == OPTION_VOLUME)
		{
		    int			i;
		    BUF			buf;
		    // Volume
		    glbChooser->clear();

		    for (i = 0; i <= 10; i++)
		    {
			buf.sprintf("%3d%% Volume", (10 - i) * 10);

			glbChooser->appendChoice(buf);
		    }
		    glbChooser->setChoice(10 - glbConfig->myMusicVolume);

		    while (!TCODConsole::isWindowClosed())
		    {
			redrawWorld();
			key = gfx_getKey(false);

			glbChooser->processKey(key);

			setMusicVolume(10 - glbChooser->getChoice());
			if (key)
			{
			    break;
			}
		    }
		    buildOptionsMenu(OPTION_VOLUME);
		}
#endif
		else if (glbChooser->getChoice() == OPTION_COINS)
		{
		    glbConfig->myCoinsDynamic = !glbConfig->myCoinsDynamic;
		    glbCoinStack->setBarGraph(!glbConfig->myCoinsDynamic);
		}
		else if (glbChooser->getChoice() == OPTION_DIFFICULTY)
		{
		    glbChooser->clear();

		    glbChooser->setPrequelAttr(ATTR_WHITE);
		    glbChooser->setPrequel("\nWhat style of game do you wish to play?\n\nNote: Changing restarts the game.\n\n");

		    DIFFICULTY_NAMES	difficulty;
		    FOREACH_DIFFICULTY(difficulty)
		    {
			glbChooser->appendChoice(glb_difficultydefs[difficulty].name);
		    }
		    glbChooser->setChoice( glbMap ? glbMap->getDifficulty() : DIFFICULTY_NORMAL );
		    while (!TCODConsole::isWindowClosed())
		    {
			redrawWorld();
			key = gfx_getKey(false);
			glbChooser->processKey(key);

			if (key == '\n' || key == ' ')
			{
			    // Select!
			    if (!glbMap || glbChooser->getChoice() != glbMap->getDifficulty())
				glbEngine->queue().append(COMMAND(ACTION_RESTART, 1, glbChooser->getChoice()));
			    break;
			}
			else if (key == '\x1b')
			{
			    // escape
			    break;
			}
		    }
		    buildOptionsMenu(OPTION_DIFFICULTY);
		}
		else if (glbChooser->getChoice() == OPTION_FULLSCREEN)
		{
		    glbConfig->myFullScreen = !glbConfig->myFullScreen;
		    // This is intentionally unrolled to work around a
		    // bool/int problem in libtcod
		    if (glbConfig->myFullScreen)
			TCODConsole::setFullscreen(true);
		    else
			TCODConsole::setFullscreen(false);
		}
		else if (glbChooser->getChoice() == OPTION_QUIT)
		{
		    // Quit
		    done = true;
		    break;
		}
	    }
	    else if (key == '\x1b')
	    {
		// Esc on options is to go back to play.
		// Play...
		break;
	    }
	    else
	    {
		// Ignore other options.
	    }
	}
    }
    glbChooserActive = false;

    return done;
}

bool
reloadLevel(bool alreadybuilt)
{
    if (TCODConsole::isWindowClosed())
	return true;

    glbVeryFirstRun = false;

    glbLevel = 1;
    msg_newturn();
    glbMessagePanel->clear();
    glbMessagePanel->appendText("> ");
    if (!alreadybuilt)
	glbEngine->queue().append(COMMAND(ACTION_RESTART, glbLevel,
				    glbMap ? glbMap->getDifficulty()
					   : DIFFICULTY_NORMAL));

    startMusic();

    // Wait for the map to finish.
    if (glbEngine->awaitRebuild())
    {
	popupText(text_lookup("welcome", "Back"));
    }
    else
    {
	popupText(text_lookup("welcome", "You1"));
    }

    // Redrawing the world updates our glbMap so everyone will see it.
    redrawWorld();

    return false;
}

// Let the avatar watch the end...
void
deathCoolDown()
{
    int		coolstartms = TCOD_sys_elapsed_milli();
    int		lastenginems = coolstartms;
    int		curtime;
    bool	done = false;
    double	ratio;

    while (!done)
    {
	curtime = TCOD_sys_elapsed_milli();

	// Queue wait events, one per 100 ms.
	while (lastenginems + 200 < curtime)
	{
	    glbEngine->queue().append(COMMAND(ACTION_WAIT));
	    lastenginems += 200;
	}

	// Set our fade.
	ratio = (curtime - coolstartms) / 5000.;
	ratio = 1.0 - ratio;
	if (ratio < 0)
	{
	    done = true;
	    ratio = 0;
	}

	TCOD_console_set_fade( (u8) (255 * ratio), TCOD_black);

	// Keep updating
	redrawWorld();
    }

    gfx_clearKeyBuf();

    TCOD_console_set_fade(255,  TCOD_black);
}

void
shutdownEverything()
{
    glbConfig->save("../smith.cfg");

    // Save the game.  A no-op if dead.
    glbEngine->queue().append(COMMAND(ACTION_SAVE));
    glbEngine->awaitSave();

    stopMusic(true);
#ifdef USE_AUDIO
    if (glbTunes)
	Mix_FreeMusic(glbTunes);
#endif

    gfx_shutdown();

#ifdef USE_AUDIO
    SDL_QuitSubSystem(SDL_INIT_AUDIO);
#endif

    SDL_Quit();
}

void
queryOfferPrice(COMMAND &cmd)
{
    MOB		*avatar = (glbMap ? glbMap->avatar() : 0);
    if (!avatar || !avatar->alive())
    {
	cmd.action() = ACTION_NONE;
	return;
    }

    MOB		*at = avatar->map()->findMob(cmd.dy());
    ITEM	*item = avatar->getItemFromId(cmd.dz());
    if (!item || !at)
    {
	cmd.action() = ACTION_NONE;
	return;
    }

    BUF		query;
    BUF		itemname = item->getSingleName();

    query.sprintf("You notice %s has %d gold.\nHow much will you demand for %s?\n",
	    at->getName(),
	    at->getGold(),
	    itemname.buffer() );

    BUF		amount = popupQuestion(query);

    cmd.dw() = atoi(amount.buffer());
    if (cmd.dw() <= 0)
	cmd.action() = ACTION_NONE;
}

void
queryPurchasePrice(COMMAND &cmd)
{
    MOB		*avatar = (glbMap ? glbMap->avatar() : 0);
    if (!avatar || !avatar->alive())
    {
	cmd.action() = ACTION_NONE;
	return;
    }

    MOB		*at = avatar->map()->findMob(cmd.dy());
    ITEM	*item = at->getItemFromId(cmd.dz());

    if (!item || !at)
    {
	cmd.action() = ACTION_NONE;
	return;
    }
    BUF		query;
    BUF		itemname = item->getSingleName();

    query.sprintf("You have %d gold.\nHow much will you offer for %s?\n",
	    avatar->getGold(),
	    itemname.buffer() );

    BUF		amount = popupQuestion(query);

    cmd.dw() = atoi(amount.buffer());
    if (cmd.dw() <= 0)
	cmd.action() = ACTION_NONE;
}

void
buildShopMenu(MOB_NAMES keeper, PTRLIST<COMMAND> &list, PTRLIST<int> &hotkeys)
{
    glbChooser->clear();
    list.clear();
    BUF			buf;
    MOB		*avatar = (glbMap ? glbMap->avatar() : 0);
    bool	valid = false;

    if (!avatar || !avatar->alive())
	return;

    glbChooser->setPrequelAttr((ATTR_NAMES) glb_mobdefs[keeper].attr);
    switch (keeper)
    {
	case MOB_HEALER:
	{
	    int		hpneed;

	    glbChooser->setPrequel(text_lookup("shop", "healer"));

	    hpneed = avatar->getMaxHP() - avatar->getHP();

	    if (avatar->hasItem(ITEM_POISON))
	    {
		buf.sprintf("%d gold - Cure poison", 10);
		glbChooser->appendChoice(buf);
		list.append(COMMAND(ACTION_SHOP, SHOP_REMOVEITEM, 
				    10, ITEM_POISON));
	    }
	    // Can the player even get here while still blind?
	    if (avatar->hasItem(ITEM_BLIND))
	    {
		buf.sprintf("%d gold - Cure blindness", 20);
		glbChooser->appendChoice(buf);
		list.append(COMMAND(ACTION_SHOP, SHOP_REMOVEITEM, 
				    10, ITEM_BLIND));
	    }

	    valid = true;
	    if (hpneed)
	    {
		buf.sprintf("%d gold - Full recovery", hpneed);
		glbChooser->appendChoice(buf);
		list.append(COMMAND(ACTION_SHOP, SHOP_HEAL, hpneed, hpneed));
	    }
	    hpneed /= 2;
	    if (hpneed)
	    {
		buf.sprintf("%d gold - Half recovery", hpneed);
		glbChooser->appendChoice(buf);
		list.append(COMMAND(ACTION_SHOP, SHOP_HEAL, hpneed, hpneed));
	    }
	    hpneed /= 2;
	    if (hpneed)
	    {
		buf.sprintf("%d gold - Quarter recovery", hpneed);
		glbChooser->appendChoice(buf);
		list.append(COMMAND(ACTION_SHOP, SHOP_HEAL, hpneed, hpneed));
	    }

	    glbChooser->appendChoice("0 gold - A few moments peace and quiet");
	    list.append(COMMAND(ACTION_SHOP, SHOP_HEAL, 0, 0));

	    break;
	}
	case MOB_ELDER:
	{
	    glbChooser->setPrequel(text_lookup("shop", "elder"));
	    valid = true;

	    glbChooser->appendChoice("Tell me of the cave?");
	    list.append(COMMAND(ACTION_SHOP, SHOP_GOSSIP, 0, GOSSIP_INTRO));

	    glbChooser->appendChoice("Tell me of the isle?");
	    list.append(COMMAND(ACTION_SHOP, SHOP_GOSSIP, 0, GOSSIP_MEDITATE));

	    glbChooser->appendChoice("Tell me of equipment?");
	    list.append(COMMAND(ACTION_SHOP, SHOP_GOSSIP, 0, GOSSIP_EQUIP));

	    // Identification requests
	    for (int itemno = 0; itemno < avatar->inventory().entries(); itemno++)
	    {
		ITEM	*item = avatar->inventory()(itemno);

		if (item->isPotion() && !item->isMagicClassKnown())
		{
		    BUF itemname;
		    itemname = item->getName();
		    buf.sprintf("%d gold - Identify %s",
				10,
				itemname.buffer());
		    glbChooser->appendChoice(buf.buffer());
		    list.append(COMMAND(ACTION_SHOP, SHOP_ID, 10, itemno));
		}
		if (item->isRing() && !item->isMagicClassKnown())
		{
		    BUF itemname;
		    itemname = item->getName();
		    buf.sprintf("%d gold - Identify %s",
				20,
				itemname.buffer());
		    glbChooser->appendChoice(buf.buffer());
		    list.append(COMMAND(ACTION_SHOP, SHOP_ID, 20, itemno));
		}
	    }
	    break;
	}
	case MOB_MERCHANT:
	{
	    glbChooser->setPrequel(text_lookup("shop", "merchant"));
	    valid = true;

	    int			curhotkey = 0;
	    int			avatargold = avatar->getGold();
	    // Ingot requests.
	    MATERIAL_NAMES	material;
	    FOREACH_MATERIAL(material)
	    {
		if (material == MATERIAL_NONE)
		    continue;

		int		cost = glb_materialdefs[material].cost;

		buf.sprintf("[%c] 1 %s ingot - %d gold",
			    curhotkey + 'A',
			    glb_materialdefs[material].name,
			    cost);
		glbChooser->appendChoice(buf.buffer(),
					cost <= avatargold ? 0 : 100);
		list.append(COMMAND(ACTION_SHOP, SHOP_INGOTS, cost, material));
		hotkeys.append(curhotkey+'a');
		curhotkey++;
		// Skip movement and 'i'
		if (curhotkey == 8)
		    curhotkey += 3;

		buf.sprintf("[%c] 5 %s ingots - %d gold",
			    curhotkey + 'A',
			    glb_materialdefs[material].name,
			    cost*5);
		glbChooser->appendChoice(buf.buffer(),
					cost*5 <= avatargold ? 0 : 100);
		list.append(COMMAND(ACTION_SHOP, SHOP_INGOTS, cost*5, material));
		hotkeys.append(curhotkey+'a');
		curhotkey++;
		// Skip movement and 'i'
		if (curhotkey == 8)
		    curhotkey += 3;
	    }
	    // Forges.
	    if (glbMap)
	    {
		POS	p = glbMap->getRandomTile(TILE_FUTURE_FORGE);
		if (p.valid())
		{
		    int	nforge = glbMap->getMobCount(MOB_FORGE);

		    buf.sprintf("[%c] A new forge - %d gold",
			    curhotkey + 'A',
			    10 << nforge);
		    glbChooser->appendChoice(buf.buffer(),
					(10 << nforge) <= avatargold ? 0 : 100);
		    list.append(COMMAND(ACTION_SHOP, SHOP_BUYFORGE, 10 << nforge));
		    hotkeys.append(curhotkey+'a');
		    curhotkey++;
		    // Skip movement and 'i'
		    if (curhotkey == 8)
			curhotkey += 3;
		}
	    }
	    break;
	}
	case MOB_SMITH:
	{
	    valid = true;
	    glbChooser->setPrequel(text_lookup("shop", "smith"));

	    for (int itemno = 0; itemno < avatar->inventory().entries(); itemno++)
	    {
		ITEM	*item = avatar->inventory()(itemno);

		if ((item->isMelee() || item->isArmour()) && item->isBroken())
		{
		    BUF itemname;
		    itemname = item->getName();
		    buf.sprintf("%d gold - Repair %s",
				item->defn().repaircost,
				itemname.buffer());
		    glbChooser->appendChoice(buf.buffer());
		    list.append(COMMAND(ACTION_SHOP, SHOP_FIX, item->defn().repaircost, itemno));
		}
	    }

	    break;
	}
	case MOB_FLETCHER:
	{
	    valid = true;
	    glbChooser->setPrequel(text_lookup("shop", "fletcher"));

	    // ALways sell arrows
	    glbChooser->appendChoice("6 gold - A dozen arrows");
	    list.append(COMMAND(ACTION_SHOP, SHOP_BUY, 6, ITEM_ARROW));

	    for (int itemno = 0; itemno < avatar->inventory().entries(); itemno++)
	    {
		ITEM	*item = avatar->inventory()(itemno);

		if (item->isRanged() && item->isBroken())
		{
		    BUF itemname;
		    itemname = item->getName();
		    buf.sprintf("%d gold - Repair %s",
				item->defn().repaircost,
				itemname.buffer());
		    glbChooser->appendChoice(buf.buffer());
		    list.append(COMMAND(ACTION_SHOP, SHOP_FIX, item->defn().repaircost, itemno));
		}
	    }

	    break;
	}
    }


    if (valid)
    {
	glbChooser->appendChoice("No thank you.");
	list.append(COMMAND(ACTION_NONE));
	glbChooser->setChoice(list.entries()-1);
	glbChooserActive = true;
    }
    else
    {
	glbChooser->clear();
	list.clear();
    }
}

BUF
describeAdventurer(MOB *seller)
{
    BUF 		buf;

    ITEM		*weap = seller->lookupWeapon();
    ITEM		*arm = seller->lookupArmour();
    BUF		weapname;
    weapname.strcat("none");
    if (weap)
	weapname = weap->getName();
    BUF		armname;
    armname.strcat("none");
    if (arm)
	armname = arm->getName();

    buf.sprintf("You examine %s:\n\nHealth: %d / %d\nGold:   %d\nKills:  %d\nWeapon: %s (%d)\nArmour: %s (%d)\n",
		seller->getName(),
		seller->getHP(),
		seller->getMaxHP(),
		seller->getGold(),
		seller->getKills(),
		weapname.buffer(),
		weap ? weap->getKillSpeed() : 0,
		armname.buffer(),
		arm ? arm->getAC() : 100
		);
    return buf;
}

void
buildAdventurerSellMenu(MOB *seller, PTRLIST<COMMAND> &list, PTRLIST<int> &hotkeys)
{
    glbChooser->clear();
    list.clear();
    BUF			buf;
    MOB		*avatar = (glbMap ? glbMap->avatar() : 0);

    if (!avatar || !avatar->alive())
	return;

    glbChooser->setPrequelAttr((ATTR_NAMES) glb_mobdefs[MOB_ADVENTURER].attr);

    BUF			prequel;

    const char *greeting[] = 
    {
	"vendor sell",
	"I have wrested these items from beings most foul, I expect a good price!",
	"Perhaps these items will be of more use to you?",
	"What can I get for these?",
	0
    };

    const char *actualgreeting = rand_string(greeting);

    buf.sprintf("%s approaches you, saying: \"%s\"\n",
		seller->getName(),
		actualgreeting);

    seller->doShout(actualgreeting);
    prequel.strcat("\n");
    prequel.strcat(buf);
    prequel.strcat("\n");

    prequel.strcat(describeAdventurer(seller));
    prequel.strcat("\n");

    prequel = prequel.wordwrap(50);

    glbChooser->setPrequel(prequel);

    int curhotkey = 0;
    for (int i = 0; i < seller->inventory().entries(); i++)
    {
	ITEM	*ci = seller->inventory()(i);
	bool	offer = false;

	// Anything not equipped is junk to us!
	if (!ci->isEquipped())
	    offer = true;
	// Rather silly to sell gold!
	if (ci->getDefinition() == ITEM_GOLD)
	    continue;

	if (offer)
	{
	    BUF itemname;
	    itemname = ci->getName();
	    buf.sprintf("[%c] Buy %s",
			curhotkey + 'A',
			itemname.buffer());
	    glbChooser->appendChoice(buf.buffer());
	    list.append(COMMAND(ACTION_SHOP, SHOP_BUYFROM, seller->getUID(), ci->getUID()));
	    hotkeys.append(curhotkey+'a');
	    curhotkey++;
	    // Skip movement and 'i'
	    if (curhotkey == 8)
		curhotkey += 3;
	}
    }

    glbChooser->appendChoice("[Z] You have nothing of interest to me.");
    hotkeys.append('z');
    list.append(COMMAND(ACTION_SHOP, SHOP_REJECT, seller->getUID()));
    glbChooser->setChoice(list.entries()-1);
    glbChooserActive = true;
}

void
buildAdventurerBuyMenu(MOB *seller, PTRLIST<COMMAND> &list, PTRLIST<int> &hotkeys)
{
    glbChooser->clear();
    list.clear();
    BUF			buf;
    MOB		*avatar = (glbMap ? glbMap->avatar() : 0);

    if (!avatar || !avatar->alive())
	return;

    glbChooser->setPrequelAttr((ATTR_NAMES) glb_mobdefs[MOB_ADVENTURER].attr);

    BUF			prequel;

    const char *greeting[] = 
    {
	"vendor buy",
	"I am in need of fine weapons and armour!",
	"What exciting metal products do you have in stock?",
	"May I see your wares?",
	0
    };

    const char *actualgreeting = rand_string(greeting);

    buf.sprintf("%s approaches you, saying: \"%s\"\n",
		seller->getName(),
		actualgreeting);

    seller->doShout(actualgreeting);
    prequel.strcat("\n");
    prequel.strcat(buf);
    prequel.strcat("\n");

    prequel.strcat(describeAdventurer(seller));
    prequel.strcat("\n");

    prequel = prequel.wordwrap(50);
    glbChooser->setPrequel(prequel);

    // List all items that the adventurer likes more!
    ITEM		*cur_weapon, *cur_armour;

    cur_weapon = seller->lookupWeapon();
    cur_armour = seller->lookupArmour();

    ITEMLIST		sellitems;
    for (int i = 0; i < avatar->inventory().entries(); i++)
    {
	ITEM	*ci = avatar->inventory()(i);

	if (ci->isEquipped())
	    continue;
	if (ci->isArmour())
	{
	    if (seller->aiLikeMoreArmour(cur_armour, ci) == ci)
	    {
		sellitems.append(ci);
		continue;
	    }
	}
	if (ci->isMelee())
	{
	    if (seller->aiLikeMoreWeapon(cur_weapon, ci) == ci)
	    {
		sellitems.append(ci);
		continue;
	    }
	}
    }

    ITEMPTR_QUALITYSORT op;
    sellitems.stableSort(op);

    int curhotkey = 0;
    for (int i = 0; i < sellitems.entries(); i++)
    {
	ITEM	*ci = sellitems(i);

	BUF itemname;
	itemname = ci->getSingleName();
	buf.sprintf("[%c] Sell %s (%d %s)",
		    curhotkey + 'A',
		    itemname.buffer(),
		    glb_itemdefs[ci->getDefinition()].killspeed 
			? ci->getKillSpeed()
			: ci->getAC(),
		    glb_itemdefs[ci->getDefinition()].killspeed 
			? "power"
			: "armour"
		    );
	glbChooser->appendChoice(buf.buffer());
	list.append(COMMAND(ACTION_SHOP, SHOP_SELLTO, seller->getUID(), ci->getUID()));
	hotkeys.append(curhotkey+'a');
	curhotkey++;
	// Skip movement and 'i'
	if (curhotkey == 8)
	    curhotkey += 3;
    }

    // Always give the failure+
    glbChooser->appendChoice("[Z] I have nothing to offer you.");
    list.append(COMMAND(ACTION_SHOP, SHOP_REJECT, seller->getUID()));
    hotkeys.append('z');
    glbChooser->setChoice(list.entries()-1);
    glbChooserActive = true;
}

void
buildForgeActionMenu(PTRLIST<SHOP_NAMES> &list, PTRLIST<int> &hotkeys)
{
    glbChooser->clear();
    list.clear();
    BUF			buf, prequel;

    glbChooser->setPrequelAttr(ATTR_FIRE);

    prequel.strcat("\nWhat shall you do at the forge?\n\n");

    prequel = prequel.wordwrap(50);
    glbChooser->setPrequel(prequel);

    // List all that can be made!
    glbChooser->appendChoice("[M] Melt Item into Ingots");
    list.append(SHOP_SMELT);
    hotkeys.append('m');
    glbChooser->appendChoice("[C] Craft Item from Ingots");
    list.append(SHOP_FORGE);
    hotkeys.append('c');

    // Always give the failure+
    glbChooser->appendChoice("[Esc] Cancel");
    list.append(SHOP_NONE);
    glbChooser->setChoice(list.entries()-1);
    glbChooserActive = true;
}

SHOP_NAMES
chooseForgeAction()
{
    PTRLIST<SHOP_NAMES>	list;
    PTRLIST<int>	hotkeys;

    buildForgeActionMenu(list, hotkeys);
    return runMenuSelection(list, SHOP_NONE, hotkeys);
}

void
buildMaterialMenu(PTRLIST<MATERIAL_NAMES> &list, PTRLIST<int> &hotkeys)
{
    glbChooser->clear();
    list.clear();
    BUF			buf;
    MOB		*avatar = (glbMap ? glbMap->avatar() : 0);

    if (!avatar || !avatar->alive())
	return;

    glbChooser->setPrequelAttr(ATTR_FIRE);

    BUF			prequel;

    prequel.strcat("\nWhat material will you use?\n\n");

    prequel = prequel.wordwrap(50);
    glbChooser->setPrequel(prequel);

    // List all that can be made!
    MATERIAL_NAMES	material;
    FOREACH_MATERIAL(material)
    {
	if (material == MATERIAL_NONE)
	    continue;
	BUF	materialname = gram_capitalize(glb_materialdefs[material].name);

	buf.sprintf("[%c] %s (%d ingot%s)", 
		    materialname.buffer()[0],
		    materialname.buffer(),
		    avatar->getIngots(material),
		    avatar->getIngots(material)==1 ? "" : "s");
	glbChooser->appendChoice(buf.buffer());
	hotkeys.append(glb_materialdefs[material].name[0]);
	list.append(material);
    }

    // Always give the failure+
    glbChooser->appendChoice("[Esc] Cancel");
    list.append(MATERIAL_NONE);
    glbChooser->setChoice(list.entries()-1);
    glbChooserActive = true;
}

MATERIAL_NAMES
chooseForgeMaterial()
{
    PTRLIST<MATERIAL_NAMES>	list;
    PTRLIST<int>		hotkeys;

    buildMaterialMenu(list, hotkeys);
    return runMenuSelection(list, MATERIAL_NONE, hotkeys);
}

void
buildForgeMenu(MOB *forge, PTRLIST<COMMAND> &list, PTRLIST<int> &hotkeys, MATERIAL_NAMES material)
{
    glbChooser->clear();
    list.clear();
    BUF			buf;
    MOB		*avatar = (glbMap ? glbMap->avatar() : 0);

    if (!avatar || !avatar->alive())
	return;

    glbChooser->setPrequelAttr(ATTR_FIRE);

    BUF			prequel;

    prequel.strcat("\nWhat do you wish to make?\n\n");
    buf.sprintf("You have %d %s ingot%s.\n\n",
		    avatar->getIngots(material),
		    glb_materialdefs[material].name,
		    avatar->getIngots(material)==1 ? "" : "s");
    prequel.strcat(buf);

    prequel = prequel.wordwrap(50);
    glbChooser->setPrequel(prequel);

    // List all that can be made!

    int			curhotkey = 0;
    ITEM_NAMES		item;
    FOREACH_ITEM(item)
    {
	if (glb_itemdefs[item].ingots)
	{
	    BUF itemname, iname, mname;
	    iname = gram_capitalize(glb_itemdefs[item].name);
	    mname = gram_capitalize(glb_materialdefs[material].name);
	    itemname.sprintf("[%c] %s %s (%d %s, %d ingot%s, %d turns)", 
				curhotkey + 'A',
				mname.buffer(),
				iname.buffer(),
				glb_itemdefs[item].killspeed 
				    ? ITEM::getKillSpeed(item, material) 
				    : ITEM::getAC(item, material),
				glb_itemdefs[item].killspeed 
				    ? "Power"
				    : "Armour",
				glb_itemdefs[item].ingots,
				glb_itemdefs[item].ingots==1 ? "" : "s",
				ITEM::getCraftTime(item, material));
	    // Set the percentage depending if we can craft it.
	    glbChooser->appendChoice(itemname,
		    avatar->getIngots(material) >= glb_itemdefs[item].ingots ? 0 : 100);
	    list.append(COMMAND(ACTION_SHOP, SHOP_FORGE, forge->getUID(), item, material));
	    hotkeys.append(curhotkey+'a');
	    curhotkey++;
	    // Skip movement and 'i'
	    if (curhotkey == 8)
		curhotkey += 3;
	}
    }

    // Always give the failure+
    glbChooser->appendChoice("[Esc] Make nothing");
    list.append(COMMAND(ACTION_NONE));
    glbChooser->setChoice(list.entries()-1);
    glbChooserActive = true;
}

void
buildSmeltMenu(MOB *forge, PTRLIST<COMMAND> &list, PTRLIST<int> &hotkeys)
{
    glbChooser->clear();
    list.clear();
    BUF			buf;
    MOB		*avatar = (glbMap ? glbMap->avatar() : 0);

    if (!avatar || !avatar->alive())
	return;

    glbChooser->setPrequelAttr(ATTR_FIRE);

    BUF			prequel;

    prequel.strcat("\nWhat do you wish to melt down?\n\n");
    prequel.strcat(buf);

    prequel = prequel.wordwrap(50);
    glbChooser->setPrequel(prequel);

    int			curhotkey = 0;

    // List all that can be made!
    ITEMLIST		items = avatar->inventory();
    // Sort by broken!
    ITEMPTR_QUALITYSORT	op;
    items.stableSort(op);

    for (int i = 0; i < items.entries(); i++)
    {
	ITEM	*ci = items(i);

	if (ci->isEquipped())
	    continue;
	if (ci->material() == MATERIAL_NONE)
	    continue;
	if (ci->getDefinition() == ITEM_INGOT)
	    continue;

	BUF itemname = ci->getSingleName();

	buf.sprintf("[%c] %sMelt %s into %d ingot%s (%d turns)",
		curhotkey + 'A',
		ci->isBroken() ? "* " : "",
		itemname.buffer(),
		ci->defn().ingots,
		ci->defn().ingots==1?"":"s",
		ci->getCraftTime() / 4);
	glbChooser->appendChoice(buf.buffer());
	list.append(COMMAND(ACTION_SHOP, SHOP_SMELT, forge->getUID(), ci->getUID()));
	hotkeys.append(curhotkey+'a');
	curhotkey++;
	// Skip movement and 'i'
	if (curhotkey == 8)
	    curhotkey += 3;
    }

    // Always give the failure+
    glbChooser->appendChoice("[Esc] Cancel");
    list.append(COMMAND(ACTION_NONE));
    glbChooser->setChoice(list.entries()-1);
    glbChooserActive = true;
}

void
runShop(int shopkeeper_uid)
{
    PTRLIST<COMMAND>	shoppinglist;
    PTRLIST<int>	hotkeys;

    if (!glbMap)
	return;
    MOB		*shopkeeper_mob = glbMap->findMob(shopkeeper_uid);
    if (!shopkeeper_mob)
	return;
    // Now see if this is an adventurer, in which case it is the
    // other way around!
    MOB_NAMES		shopkeeper = shopkeeper_mob->getDefinition();

    if (shopkeeper == MOB_ADVENTURER)
    {
	if (shopkeeper_mob->getStrategy() == STRATEGY_BUY)
	{
	    buildAdventurerBuyMenu(shopkeeper_mob, shoppinglist, hotkeys);
	}
	else
	{
	    buildAdventurerSellMenu(shopkeeper_mob, shoppinglist, hotkeys);
	}
    }
    else if (shopkeeper == MOB_FORGE)
    {
	while (1)
	{
	    SHOP_NAMES	action = chooseForgeAction();

	    if (action == SHOP_NONE)
		return;
	    if (action == SHOP_FORGE)
	    {
		while (1)
		{
		    MATERIAL_NAMES	material = chooseForgeMaterial();
		    if (material == MATERIAL_NONE)
			break;

		    // Any call into chooser must re-fetch the shopkeeper
		    shopkeeper_mob = glbMap->findMob(shopkeeper_uid);
		    if (!shopkeeper_mob)
			return;

		    buildForgeMenu(shopkeeper_mob, shoppinglist, hotkeys, material);

		    // Run the chooser...
		    COMMAND	cmd = runMenuSelection(shoppinglist, COMMAND(ACTION_NONE), hotkeys);

		    if (cmd.action() != ACTION_NONE)
		    {
			glbEngine->queue().append(cmd);
			return;
		    }
		}
	    }
	    else if (action == SHOP_SMELT)
	    {
		// Any call into chooser must re-fetch the shopkeeper
		shopkeeper_mob = glbMap->findMob(shopkeeper_uid);
		if (!shopkeeper_mob)
		    return;

		buildSmeltMenu(shopkeeper_mob, shoppinglist, hotkeys);
		// Run the chooser...
		COMMAND	cmd = runMenuSelection(shoppinglist, COMMAND(ACTION_NONE), hotkeys);

		if (cmd.action() != ACTION_NONE)
		{
		    glbEngine->queue().append(cmd);
		    return;
		}
	    }
	}
    }
    else
    {
	buildShopMenu(shopkeeper, shoppinglist, hotkeys);
    }
    
    if (shoppinglist.entries() == 0)
	return;

    // Run the chooser...
    COMMAND	cmd = runMenuSelection(shoppinglist, COMMAND(ACTION_NONE), hotkeys);

    if (cmd.action() == ACTION_SHOP && cmd.dx() == SHOP_SELLTO)
    {
	queryOfferPrice(cmd);
    }
    if (cmd.action() == ACTION_SHOP && cmd.dx() == SHOP_BUYFROM)
    {
	queryPurchasePrice(cmd);
    }
    if (cmd.action() != ACTION_NONE)
    {
	glbEngine->queue().append(cmd);
    }
}

#ifdef WIN32
int WINAPI
WinMain(HINSTANCE hINstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCMdSHow)
#else
int 
main(int argc, char **argv)
#endif
{
    bool		done = false;

    // Clamp frame rate.
    TCOD_sys_set_fps(60);

    glbConfig = new CONFIG();
    glbConfig->load("../smith.cfg");

    // Dear Microsoft,
    // The following code in optimized is both a waste and seems designed
    // to ensure that if we call a bool where the callee pretends it is
    // an int we'll end up with garbage data.
    //
    // 0040BF4C  movzx       eax,byte ptr [eax+10h] 

    // TCODConsole::initRoot(SCR_WIDTH, SCR_HEIGHT, "The Smith's Hand", myFullScreen);
    // 0040BF50  xor         ebp,ebp 
    // 0040BF52  cmp         eax,ebp 
    // 0040BF54  setne       cl   
    // 0040BF57  mov         dword ptr [esp+28h],eax 
    // 0040BF5B  push        ecx  

    // My work around is to constantify the fullscreen and hope that
    // the compiler doesn't catch on.
    if (glbConfig->myFullScreen)
	TCODConsole::initRoot(SCR_WIDTH, SCR_HEIGHT, "The Smith's Hand", true);
    else
	TCODConsole::initRoot(SCR_WIDTH, SCR_HEIGHT, "The Smith's Hand", false);

    // TCOD doesn't do audio.
#ifdef USE_AUDIO
    SDL_InitSubSystem(SDL_INIT_AUDIO);
    if (Mix_OpenAudio(22050, AUDIO_S16, 2, 4096))
    {
	printf("Failed to open audio!\n");
	exit(1);
    }
#endif

    SDL_InitSubSystem(SDL_INIT_JOYSTICK);

    setMusicVolume(glbConfig->musicVolume());

    rand_setseed((long) time(0));

    gfx_init();
    MAP::init();

    text_init();

    glbDisplay = new DISPLAY(0, 0, 80, 40);

    glbMessagePanel = new PANEL(40, 10);
    msg_registerpanel(glbMessagePanel);
    glbMessagePanel->move(20, 40);

    glbPopUp = new PANEL(50, 30);
    glbPopUp->move(15, 10);
    glbPopUp->setBorder(true, ' ', ATTR_BORDER);
    // And so the mispelling is propagated...
    glbPopUp->setRigthMargin(1);
    glbPopUp->setIndent(1);

    glbJacob = new PANEL(30, 25);
    glbJacob->move(55, 2);
    glbJacob->setTransparent(true);

    glbWeaponInfo = new PANEL(30, 30);
    glbWeaponInfo->move(1, 2);
    glbWeaponInfo->setTransparent(true);

    glbVictoryInfo = new PANEL(26, 5);
    glbVictoryInfo->move(28, 5);
    glbVictoryInfo->setIndent(2);
    glbVictoryInfo->setBorder(true, ' ', ATTR_VICTORYBORDER);
    glbVictoryInfo->clear();
    glbVictoryInfo->newLine();
    glbVictoryInfo->appendText("The orcs are all dead!");
    glbVictoryInfo->newLine();
    glbVictoryInfo->newLine();
    glbVictoryInfo->appendText("Press [V] to finish.");

    glbLevel = 1;

    glbChooser = new CHOOSER();
    glbChooser->move(40, 25, CHOOSER::JUSTCENTER, CHOOSER::JUSTCENTER);
    glbChooser->setBorder(true, ' ', ATTR_BORDER);
    glbChooser->setIndent(1);
    glbChooser->setRigthMargin(1);
    glbChooser->setAttr(ATTR_NORMAL, ATTR_HILITE, ATTR_LIGHTBLACK, ATTR_DKHILITE);

    glbLevelBuildStatus = new CHOOSER();
    glbLevelBuildStatus->move(40, 25, CHOOSER::JUSTCENTER, CHOOSER::JUSTCENTER);
    glbLevelBuildStatus->setBorder(true, ' ', ATTR_BORDER);
    glbLevelBuildStatus->setIndent(1);
    glbLevelBuildStatus->setRigthMargin(1);
    glbLevelBuildStatus->setAttr(ATTR_NORMAL, ATTR_HILITE, ATTR_HILITE,
				ATTR_HILITE);

    glbEngine = new ENGINE(glbDisplay);

    glbHealthFire = new FIREFLY(0, 80, 1, FIRE_HEALTH, true);
    glbHealthFire->setBarGraph(true);
    glbManaFire = new FIREFLY(0, 80, 1, FIRE_MANA, true);
    glbManaFire->setBarGraph(true);
    glbCoinStack = new FIREFLY(0, 20, 50, FIRE_GOLD, false);
    glbCoinStack->setBarGraph(!glbConfig->myCoinsDynamic);

    // Run our first level immediately so it can be built or 
    // loaded from disk while people wait.
    glbEngine->queue().append(COMMAND(ACTION_RESTART, 1, DIFFICULTY_NORMAL));

    done = optionsMenu();
    if (done)
    {
	shutdownEverything();
	return 0;
    }

    done = reloadLevel(true);
    if (done)
    {
	shutdownEverything();
	return 0;
    }

    do
    {
	int		key;
	int		dx, dy;

	redrawWorld();

	while (1)
	{
	    BUF		pop;

	    pop = glbEngine->getNextPopup();
	    if (!pop.isstring())
		break;

	    popupText(pop, POPUP_DELAY);
	}

	while (1)
	{
	    int	shopper;

	    shopper = glbEngine->getNextShopper();
	    if (shopper == -1)
		break;

	    runShop(shopper);
	}


	if (glbMap && glbMap->avatar() &&
	    !glbMap->avatar()->alive())
	{
	    // You have died...
	    deathCoolDown();

	    popupText(text_lookup("game", "lose"));

	    stopMusic();

	    done = optionsMenu();
	    if (done)
	    {
		shutdownEverything();
		return 0;
	    }
	    // Don't prep a reload.
	    done = reloadLevel(false);
	}
	else if (glbMap && !glbMap->avatar())
	{
	    stopMusic();
	    done = optionsMenu();
	    if (done)
	    {
		shutdownEverything();
		return 0;
	    }
	    done = reloadLevel(false);
	}

	key = gfx_getKey(false);

	if (gfx_cookDir(key, dx, dy))
	{
	    // Direction.
	    glbEngine->queue().append(COMMAND(ACTION_BUMP, dx, dy));
	}

	switch (key)
	{
	    case 'Q':
		done = true;
		break;

	    case 'R':
		done = reloadLevel(false);
		break;

	    case 'w':
		glbEngine->queue().append(COMMAND(ACTION_WAITUNTIL));
		break;

	    case 'i':
		inventoryMenu();
		break;

	    case 'a':
		popupText(text_lookup("game", "about"));
		break;

	    case 'W':
	    {
		popupText(text_lookup("welcome", "You1"));
		break;
	    }

	    case 'P':
		glbConfig->myFullScreen = !glbConfig->myFullScreen;
		// This is intentionally unrolled to work around a
		// bool/int problem in libtcod
		if (glbConfig->myFullScreen)
		    TCODConsole::setFullscreen(true);
		else
		    TCODConsole::setFullscreen(false);
		break;
	
	    case GFX_KEYF1:
	    case '?':
		popupText(text_lookup("game", "help"));
		break;

	    case 'v':
		if (glbMap && glbMap->avatar())
		{
		    if (glbMap->avatar()->alive() &&
			glbMap->getNumOrcs() == 0)
		    {
			// Traditional win.
			{
			    // Post victoyr
			    BUF		buf;
			    buf.sprintf("%d", glbMap->avatar()->getGold());

			    text_registervariable("GOLD", buf);
			    buf.sprintf("%f", glbMap->avatar()->getGold() / (float) MAX(glbMap->getDeadAdventurers(), 1));
			    text_registervariable("GOLDPER", buf);
			}
			popupText(text_lookup("game", "victory"));

			stopMusic();
			done = reloadLevel(false);
			if (!done)
			    done = optionsMenu();
		    }
		    else
		    {
			msg_report("You haven't achieved victory conditions.  ");
		    }
		}
		break;

	    case 'x':
		doExamine();
		break;

#if 0
	    case 'W':
		glbEngine->queue().append(COMMAND(ACTION_CREATEITEM, ITEM_BOW));

		break;
#endif

	    case '\x1b':
		// If meditating, abort.
		if (glbMap && glbMap->avatar() && 
		    glbMap->avatar()->alive() && glbMap->avatar()->isMeditating())
		{
		    glbEngine->queue().append(COMMAND(ACTION_MEDITATE));
		    break;
		}
		// If waiting, abort.
		if (glbMap && glbMap->avatar() && 
		    glbMap->avatar()->alive() && glbMap->avatar()->isWaiting())
		{
		    glbEngine->queue().append(COMMAND(ACTION_STOPWAITING));
		    break;
		}
		// FALL THROUGH
	    case 'O':
		done = optionsMenu();
		break;
	}
    } while (!done && !TCODConsole::isWindowClosed());

    shutdownEverything();

    return 0;
}
