/*
 * Licensed under BSD license.  See LICENCE.TXT  
 *
 * Produced by:	Jeff Lait
 *
 *      	Smith's Hand Development
 *
 * NAME:        simulator.h ( Smith's Hand, C++ )
 *
 * COMMENTS:
 *	There are two modes for this simulation.
 *	We either go turn by turn, in which case all the AI
 *	and linking has to be handled by the mobs themselves.
 *	Or we run free, waiting for the next "interesting" thing
 *	in which case we just teleport MOBs based on estimated walk times.
 */

#ifndef __simulator__
#define __simulator__

#include <iostream>
using namespace std;

#include "ptrlist.h"
#include "thread.h"
#include "glbdef.h"
#include "map.h"
#include "pq.h"

class SIMULATOR
{
public:
    class EVENT
    {
    public:
	enum TYPE
	{
	    ADVENTURER_SPAWN,
	    ADVENTURER_RETURN,
	    ADVENTURER_DIES,
	    ORC_SPAWN,
	    ORC_KILL,
	    ORC_KILL_LITE,
	    FORGE_DONE,
	};

	EVENT() {}

	EVENT(int time, TYPE type)
	{
	    myTime = time;
	    myType = type;
	    myData = -1;
	}
	EVENT(int time, TYPE type, int data)
	{
	    myTime = time;
	    myType = type;
	    myData = data;
	}

	int		time() const { return myTime; }
	TYPE	type() const { return myType; }
	int	data() const { return myData; }

	void	save(ostream &os) const;
	void	load(istream &is);

    protected:
	int		myTime;
	TYPE		myType;
	int		myData;
    };


    SIMULATOR();
    ~SIMULATOR();

    SIMULATOR		*copy() const;

    // Applies a step of the simulation.
    void		 doHeartbeat(MAP *map);

    void		 queueEvent(EVENT event);

    void		 startAdventure(MOB *mob);
    void		 escapeAdventurer(MOB *mob);

    const MOBLIST	&awayMobs() const { return myAwayMobs; }

    void		 save(ostream &os) const;
    static SIMULATOR	*load(istream &is);
protected:
    // Owned by us!
    MOBLIST		myAwayMobs;

    PQ<EVENT, int>	myEvents;
};

#endif
